import json
from pathlib import Path

from app_utils.esi_testing import EsiClientStub, EsiEndpoint


def load_test_data():
    file_path = Path(__file__).parent / "esi.json"
    with file_path.open("r", encoding="utf-8") as fp:
        return json.load(fp)


_esi_data = load_test_data()

_endpoints = []

esi_client_stub = EsiClientStub(_esi_data, endpoints=_endpoints)
esi_client_error_stub = EsiClientStub(_esi_data, endpoints=_endpoints, http_error=True)



from squads.models.groups import Pending


def add_info_to_context(request, context: dict) -> dict:
    """Add additional information to the context for the view."""
    pending_count = Pending.objects.filter(user=request.user).count()
    new_context = {
        **{
            "pending_count": pending_count,
        },
        **context,
    }
    return new_context



"""App Configuration"""

# Django
from django.apps import AppConfig

# AA Example App
from squads import __version__


# pylint: disable=unused-import, import-outside-toplevel
class SquadsConfig(AppConfig):
    """App Config"""

    default_auto_field = "django.db.models.AutoField"
    author = "Geuthur"
    name = "squads"
    label = "squads"
    verbose_name = f"Squads v{__version__}"

    def ready(self):
        import squads.signals



"""Manage views."""

from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.db import transaction
from django.shortcuts import get_object_or_404, redirect, render
from django.utils.safestring import mark_safe

from squads.forms import SquadsGroupForm
from squads.hooks import get_extension_logger
from squads.models.groups import Groups, Pending
from squads.models.memberships import Memberships

logger = get_extension_logger(__name__)


@login_required
@permission_required("squads.squad_manager")
@transaction.atomic
def create_group(request):
    if request.method == "POST":
        form = SquadsGroupForm(request.POST, request.FILES)
        if form.is_valid():
            group = form.save(commit=False)
            group.description = mark_safe(group.description)
            if not group.image:  # Überprüfen, ob das Bildfeld leer ist
                group.image = "squads/groups_images/empty.png"  # Standardbild setzen
            group.owner = request.user
            group.save()  # Save the group to the database before creating related Skills
            messages.success(request, "Squad has been created.")
            return redirect("squads:groups")
    else:
        form = SquadsGroupForm()
    return render(request, "squads/manage/create_group.html", {"form": form})


@login_required
@permission_required("squads.squad_manager")
@transaction.atomic
def accept_group(request, application_id):
    pending = get_object_or_404(Pending, application_id=application_id)
    if pending:
        Memberships.objects.create(
            group=pending.group,
            user=pending.user,
            has_required_skills=True,
            is_active=True,
        )
        Pending.objects.filter(application_id=application_id).delete()
        messages.success(request, f"{pending.user.username} has been approved.")
    else:
        messages.error(request, "%s does not exist.", application_id)

    return redirect("squads:manage_pendings")


@login_required
@permission_required("squads.squad_manager")
def decline_group(request, application_id):
    pending = get_object_or_404(Pending, application_id=application_id)
    if pending:
        Pending.objects.filter(application_id=application_id).delete()
        messages.success(request, f"{pending.user.username} has been declined.")
    else:
        messages.error(request, f"{application_id} does not exist.")

    return redirect("squads:manage_pendings")


@login_required
@permission_required("squads.squad_manager")
def manage_pendings(request):
    manage_memberships = Groups.objects.filter(owner=request.user)
    return render(
        request,
        "squads/manage/manage_pendings.html",
        {"manage_memberships": manage_memberships},
    )


@login_required
@permission_required("squads.squad_manager")
def manage_members(request):
    manage_memberships = Memberships.objects.filter(user=request.user)
    return render(
        request,
        "squads/manage/manage_members.html",
        {"manage_memberships": manage_memberships},
    )


@login_required
@permission_required("squads.squad_manager")
def delete_membership(request, application_id):
    membership = get_object_or_404(Memberships, application_id=application_id)
    if membership:
        Memberships.objects.filter(application_id=application_id).delete()
        messages.success(request, f"{membership.user.username} has been removed.")
    else:
        messages.error(request, f"{application_id} does not exist.")

    return redirect("squads:manage_members")


@login_required
@permission_required("squads.squad_manager")
def manage_groups(request):
    manage_squads = Groups.objects.filter(owner=request.user)
    return render(
        request,
        "squads/manage/manage_squads.html",
        {"manage_squads": manage_squads},
    )


@login_required
@permission_required("squads.squad_manager")
def delete_group(request, group_id):
    group = get_object_or_404(Groups, pk=group_id)
    if group:
        Groups.objects.filter(pk=group_id).delete()
        messages.success(request, f"{group.name} has been removed.")
    else:
        messages.error(request, f"{group_id} does not exist.")

    return redirect("squads:manage_groups")


@login_required
@permission_required("squads.squad_manager")
def edit_group(request, group_id):
    group_data = get_object_or_404(Groups, pk=group_id)
    if request.method == "POST":
        form = SquadsGroupForm(request.POST, request.FILES, instance=group_data)
        if form.is_valid():
            group = form.save(commit=False)
            group.description = mark_safe(group.description)
            if not group.image:
                group.image = "squads/groups_images/empty.png"
            group.save()
            messages.success(request, f"{group_data.name} has been updated.")
            return redirect("squads:manage_groups")
        messages.error(request, "Something went wrong.")
        return render(
            request,
            "squads/manage/edit_group.html",
            {"form": form, "group": group_data},
        )
    form = SquadsGroupForm(instance=group_data)
    return render(
        request, "squads/manage/edit_group.html", {"form": form, "group": group_data}
    )



"""
Memberships Model
"""

from memberaudit.models import CharacterAsset, CharacterSkillSetCheck, SkillSet

from django.contrib.auth.models import User
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.translation import gettext_lazy as _
from eveuniverse.models import EveType

from squads.hooks import get_extension_logger
from squads.models.groups import Groups

logger = get_extension_logger(__name__)


class BaseFilter(models.Model):
    """BaseFilter Model."""

    description = models.CharField(
        max_length=500,
        help_text=_("Filter Name."),
    )

    def __str__(self) -> str:
        return str(self.description)

    class Meta:
        abstract = True


class SkillSetFilter(BaseFilter):
    """SkillSetFilter."""

    skill_sets = models.ManyToManyField(
        SkillSet,
        help_text=_(
            "Users must possess the selected skill sets with at least <strong>one</strong> character."
        ),
        related_name="squads_skill_sets",
    )

    def check_filter(self, user: User):
        """Check if user Characters has required skills."""
        qs = CharacterSkillSetCheck.objects.filter(
            character__eve_character__character_ownership__user=user,
            skill_set__in=list(self.skill_sets.all()),
            failed_required_skills__isnull=True,
        )
        return qs.exists()


class AssetsFilter(BaseFilter):
    """AssetsFilter."""

    assets = models.ManyToManyField(
        EveType,
        help_text=_(
            "Users must have at least <strong>one</strong> of the selected assets with <strong>one</strong> character."
        ),
    )

    def check_filter(self, user: User):
        """Check if user Characters has at least one of the given asset."""
        return CharacterAsset.objects.filter(
            character__eve_character__character_ownership__user=user,
            eve_type__in=list(self.assets.all()),
        ).exists()


class SquadFilter(models.Model):
    """SquadFilter show up all Filters."""

    content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, editable=False
    )
    object_id = models.PositiveIntegerField(editable=False)
    filter_object = GenericForeignKey("content_type", "object_id")

    def __str__(self) -> str:
        return str(self.filter_object)

    class Meta:
        verbose_name = "Squad Filter"


class SquadGroup(models.Model):
    """Squads Group Filter."""

    group = models.OneToOneField(Groups, on_delete=models.CASCADE)
    description = models.CharField(max_length=500, default="", blank=True)
    filters = models.ManyToManyField(SquadFilter)
    enabled = models.BooleanField(default=True)  # Not Implemented yet
    last_updated = models.DateTimeField(auto_now=True)

    def __str__(self) -> str:
        return str(f"Filter Group: {self.group.name}")

    # Check Filters
    # pylint: disable=broad-exception-caught
    def run_filters(self, user: User):
        output = []
        for check in self.filters.all():
            try:
                _filter = check.filter_object
                if _filter is None:
                    logger.debug("Filter %s is None", check)
                    continue
                run_test = _filter.check_filter(user)
            except Exception as e:
                run_test = False
                logger.warning("Filter Check Failed: %s", check)
                logger.debug("Error: %s", e, exc_info=True)
            _check = {
                "desc": check.filter_object.description,
            }
            _check["result"] = run_test
            _check["filter"] = check
            output.append(_check)
        return output

    def process_filters(self, checks):
        output = True
        for check in checks:
            output = output and check.get("result", False)
        return output

    def check_user(self, user: User):
        checks = self.run_filters(user)
        output = self.process_filters(checks)
        return output



"""
Memberships Model
"""

from django.db import models
from django.utils.translation import gettext_lazy as _

from allianceauth.authentication.models import User

from squads.hooks import get_extension_logger
from squads.models.groups import Groups
from squads.view_helpers.core import generate_unique_id

logger = get_extension_logger(__name__)


class Memberships(models.Model):
    """Memberships Model store Membership Data."""

    group = models.ForeignKey(Groups, on_delete=models.CASCADE)
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    is_active = models.BooleanField(default=True)
    has_required_skills = models.BooleanField(default=False)
    joined_at = models.DateTimeField(auto_now_add=True)
    application_id = models.CharField(
        max_length=12, default=generate_unique_id, unique=True, blank=True
    )

    def __str__(self) -> str:
        return str(self.user + " in " + self.group.name)

    class Meta:
        default_permissions = ()



from django.test import TestCase

from squads.hooks import get_extension_logger


class TestTemplateTags(TestCase):
    def test_logger_fail(self):
        with self.assertRaises(TypeError):
            get_extension_logger(1234)



"""
General Model
"""

from django.db import models
from django.utils.translation import gettext_lazy as _

from squads.hooks import get_extension_logger

logger = get_extension_logger(__name__)


class General(models.Model):
    """A model defining commonly used properties and methods for Squads."""

    class Meta:
        managed = False
        default_permissions = ()
        permissions = (
            ("basic_access", "Can access this app, Squads."),
            ("squad_manager", "Can Create / Manage Squads."),
            ("squad_admin", "Can View All Squads."),
        )



"""Hook into Alliance Auth"""

# Django
# Alliance Auth
from django.utils.translation import gettext_lazy as _

from allianceauth import hooks
from allianceauth.services.hooks import MenuItemHook, UrlHook

from squads import app_settings, urls
from squads.hooks import get_extension_logger
from squads.models.filters import AssetsFilter, SkillSetFilter

logger = get_extension_logger(__name__)


class SquadsMenuItem(MenuItemHook):
    """This class ensures only authorized users will see the menu entry"""

    def __init__(self):
        super().__init__(
            f"{app_settings.SQUADS_APP_NAME}",
            "fa-solid fa-users-between-lines",
            "squads:index",
            navactive=["squads:"],
        )

    def render(self, request):
        if request.user.has_perm("squads.basic_access"):
            return MenuItemHook.render(self, request)
        return ""


@hooks.register("menu_item_hook")
def register_menu():
    """Register the menu item"""

    return SquadsMenuItem()


@hooks.register("url_hook")
def register_urls():
    """Register app urls"""

    return UrlHook(urls, "squads", r"^squads/")


@hooks.register("squads_filters")
def filters():
    return [SkillSetFilter, AssetsFilter]



# Generated by Django 4.2.11 on 2024-06-09 20:58

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

import squads.view_helpers.core


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("memberaudit", "0018_charactercloneinfo"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="General",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "permissions": (
                    ("basic_access", "Can access this app, Squads."),
                    ("squad_manager", "Can Create / Manage Squads."),
                    ("squad_admin", "Can View All Squads."),
                ),
                "managed": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Groups",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255, unique=True)),
                ("description", models.TextField(blank=True)),
                ("is_active", models.BooleanField(default=True)),
                ("require_approval", models.BooleanField(default=False)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                (
                    "image",
                    models.ImageField(
                        blank=True, null=True, upload_to="squads/groups_images/"
                    ),
                ),
                (
                    "owner",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="group_owner",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="SquadFilter",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("object_id", models.PositiveIntegerField(editable=False)),
                (
                    "content_type",
                    models.ForeignKey(
                        editable=False,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
            ],
            options={
                "verbose_name": "Squad Filter",
            },
        ),
        migrations.CreateModel(
            name="SquadGroup",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "description",
                    models.CharField(blank=True, default="", max_length=500),
                ),
                ("enabled", models.BooleanField(default=True)),
                ("last_updated", models.DateTimeField(auto_now=True)),
                ("filters", models.ManyToManyField(to="squads.squadfilter")),
                (
                    "group",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE, to="squads.groups"
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="SkillSetFilter",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "description",
                    models.CharField(
                        help_text="Filter description shown to Users.", max_length=500
                    ),
                ),
                (
                    "skill_sets",
                    models.ManyToManyField(
                        help_text="Users must possess the selected skill sets with at least <strong>one</strong> character.",
                        related_name="squads_skill_sets",
                        to="memberaudit.skillset",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Pending",
            fields=[
                ("approved", models.BooleanField(default=False)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("application", models.TextField(blank=True)),
                (
                    "application_id",
                    models.CharField(
                        default=squads.view_helpers.core.generate_unique_id,
                        editable=False,
                        max_length=12,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "group",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="squads.groups"
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Memberships",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("is_active", models.BooleanField(default=True)),
                ("has_required_skills", models.BooleanField(default=False)),
                ("joined_at", models.DateTimeField(auto_now_add=True)),
                (
                    "application_id",
                    models.CharField(
                        blank=True,
                        default=squads.view_helpers.core.generate_unique_id,
                        max_length=12,
                        unique=True,
                    ),
                ),
                (
                    "group",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="squads.groups"
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
    ]



"""Admin models"""

from typing import Any

from django.contrib import admin
from django.db.models.query import QuerySet
from django.http.request import HttpRequest

from squads.models.filters import AssetsFilter, SkillSetFilter, SquadFilter, SquadGroup


@admin.register(SkillSetFilter)
class SkillSetFilterAdmin(admin.ModelAdmin):
    """
    SkillSetFilterAdmin
    """

    list_display = ("description", "_skill_sets")
    filter_horizontal = ("skill_sets",)

    def get_queryset(self, request: HttpRequest) -> QuerySet[Any]:
        qs = super().get_queryset(request)

        return qs.prefetch_related("skill_sets")

    @admin.display()
    def _skill_sets(self, obj) -> str:
        objs = obj.skill_sets.all()

        return ", ".join(sorted([obj.name for obj in objs]))


@admin.register(AssetsFilter)
class AssetsFilterAdmin(admin.ModelAdmin):
    """
    AssetsFilterAdmin
    """

    list_display = ("description", "_assets")
    autocomplete_fields = ["assets"]

    def get_queryset(self, request: HttpRequest) -> QuerySet[Any]:
        qs = super().get_queryset(request)

        return qs.prefetch_related("assets")

    @admin.display()
    def _assets(self, obj) -> str:
        objs = obj.assets.all()

        return ", ".join(sorted([obj.name for obj in objs]))


# pylint: disable=unused-argument
@admin.register(SquadFilter)
class SquadFilterAdmin(admin.ModelAdmin):
    def has_add_permission(self, request, obj=None):
        return False

    list_display = ["__str__"]


@admin.register(SquadGroup)
class SquadGrouprAdmin(admin.ModelAdmin):
    filter_horizontal = ["filters"]
    list_display = ["__str__", "enabled"]



"""
View
"""

from .application import apply_group, leave_group
from .groups import broswe_groups, view_group
from .main import squads_index, squads_membership, squads_pending
from .manage import (
    accept_group,
    create_group,
    decline_group,
    delete_group,
    delete_membership,
    edit_group,
    manage_groups,
    manage_members,
    manage_pendings,
)



# Django
from django.contrib.auth.decorators import login_required, permission_required
from django.shortcuts import render

from squads.models.groups import Pending
from squads.models.memberships import Memberships
from squads.views._core import add_info_to_context


@login_required
@permission_required("squads.basic_access")
def squads_index(request):

    context = {}

    return render(request, "squads/index.html", add_info_to_context(request, context))


@login_required
@permission_required("squads.basic_access")
def squads_membership(request):
    memberships = Memberships.objects.filter(user=request.user)

    context = {"memberships": memberships}

    return render(
        request,
        "squads/groups/membership_groups.html",
        add_info_to_context(request, context),
    )


@login_required
@permission_required("squads.basic_access")
def squads_pending(request):
    pending_memberships = Pending.objects.filter(user=request.user, approved=False)

    context = {"pending_memberships": pending_memberships}

    return render(
        request,
        "squads/groups/pending_groups.html",
        add_info_to_context(request, context),
    )



from django.db.models.signals import post_save, pre_delete

from allianceauth import hooks

from squads.hooks import get_extension_logger
from squads.models import filters as model

logger = get_extension_logger(__name__)


class HookCache:
    all_hooks = None

    def get_hooks(self):
        if self.all_hooks is None:
            hook_array = set()
            _hooks = hooks.get_hooks("squads_filters")
            for app_hook in _hooks:
                for filter_model in app_hook():
                    if filter_model not in hook_array:
                        hook_array.add(filter_model)
            self.all_hooks = hook_array
        return self.all_hooks


filters = HookCache()


# pylint: disable=broad-exception-caught, unused-argument
def new_filter(sender, instance, created, **kwargs):
    try:
        if created:
            model.SquadFilter.objects.create(filter_object=instance)
        else:
            pass
    except Exception as e:
        logger.error("New filter failed: %s", e)


# pylint: disable=broad-exception-caught, unused-argument
def rem_filter(sender, instance, **kwargs):
    try:
        model.SquadFilter.objects.get(
            object_id=instance.pk, content_type__model=instance.__class__.__name__
        ).delete()
    except Exception as e:
        logger.error("Remove filter failed: %s", e)


for _filter in filters.get_hooks():
    post_save.connect(new_filter, sender=_filter)
    pre_delete.connect(rem_filter, sender=_filter)



"""Generate AllianceAuth test objects from allianceauth.json."""

import json
from pathlib import Path

from allianceauth.eveonline.models import (
    EveAllianceInfo,
    EveCharacter,
    EveCorporationInfo,
)


def _load_allianceauth_data():
    with open(Path(__file__).parent / "allianceauth.json", encoding="utf-8") as fp:
        return json.load(fp)


_entities_data = _load_allianceauth_data()


def load_allianceauth():
    """Load allianceauth test objects."""
    EveAllianceInfo.objects.all().delete()
    EveCorporationInfo.objects.all().delete()
    EveCharacter.objects.all().delete()
    for character_info in _entities_data.get("EveCharacter"):
        if character_info.get("alliance_id"):
            try:
                alliance = EveAllianceInfo.objects.get(
                    alliance_id=character_info.get("alliance_id")
                )
            except EveAllianceInfo.DoesNotExist:
                alliance = EveAllianceInfo.objects.create(
                    alliance_id=character_info.get("alliance_id"),
                    alliance_name=character_info.get("alliance_name"),
                    alliance_ticker=character_info.get("alliance_ticker"),
                    executor_corp_id=character_info.get("corporation_id"),
                )
        else:
            alliance = None
        try:
            corporation = EveCorporationInfo.objects.get(
                corporation_id=character_info.get("corporation_id")
            )
        except EveCorporationInfo.DoesNotExist:
            corporation = EveCorporationInfo.objects.create(
                corporation_id=character_info.get("corporation_id"),
                corporation_name=character_info.get("corporation_name"),
                corporation_ticker=character_info.get("corporation_ticker"),
                member_count=99,
                alliance=alliance,
            )
        EveCharacter.objects.create(
            character_id=character_info.get("character_id"),
            character_name=character_info.get("character_name"),
            corporation_id=corporation.corporation_id,
            corporation_name=corporation.corporation_name,
            corporation_ticker=corporation.corporation_ticker,
            alliance_id=alliance.alliance_id if alliance else None,
            alliance_name=alliance.alliance_name if alliance else "",
            alliance_ticker=alliance.alliance_ticker if alliance else "",
        )



from datetime import datetime
from typing import Optional

from ninja import Schema


class Squads(Schema):
    group: str
    group_id: int
    owner: str
    is_active: bool


class Pending(Schema):
    group: str
    group_id: int
    user: str
    approved: bool
    application: str
    application_id: str
    skills: Optional[str] = None
    created_at: datetime


class Members(Schema):
    group: str
    group_id: int
    user: str
    has_required_skills: bool
    is_active: bool
    joined_at: datetime
    application_id: str
